from loguru import logger
from argparse import ArgumentParser

from solaredge_influxdb import app

parser = ArgumentParser()
parser.add_argument(
    "--config-path",
    type=str,
    help="Path to configuration file",
)
parser.add_argument(
    "--latitude",
    type=float,
    help="Latitude of the solar panels",
)
parser.add_argument(
    "--longitude",
    type=float,
    help="Longitude of the solar panels",
)
parser.add_argument(
    "--api-key",
    type=str,
    help="API key for the SolarEdge API",
)

logger.info("Starting application; to get SolarEdge data into InfluxDB")

args = vars(parser.parse_args())
keys = list(args.keys())
for k in keys:
    if args[k] is None:
        args.pop(k)

app(**args)
logger.info("Application finished")



from .models import MeterDataResponse
from .client import SolarEdgeClient
from datetime import datetime
from loguru import logger
from typing import Union, List


class Meter(SolarEdgeClient):
    def __init__(self, api_key: str):
        super().__init__(api_key)

    def get_meters_data(
        self,
        start_time: datetime,
        end_time: datetime,
        time_unit: Union[str, None] = None,
        meters: Union[List[str], None] = None,
    ) -> Union[MeterDataResponse, None]:
        """
        Retrieves meter data from the SolarEdge API for the specified time range.

        Args:
            start_time (datetime): The start time of the data range.
            end_time (datetime): The end time of the data range.
            time_unit (str, optional): The time unit for the data aggregation. Must be one of
                "QUARTER_OF_AN_HOUR", "HOUR", "DAY", "WEEK", "MONTH", "YEAR". Defaults to None.
            meters (list[str], optional): A list of meter names to retrieve data for. Defaults to None.

        Returns:
            MeterDataResponse | None: The meter data response object if successful, None otherwise.
        """

        if time_unit:
            if time_unit not in [
                "QUARTER_OF_AN_HOUR",
                "HOUR",
                "DAY",
                "WEEK",
                "MONTH",
                "YEAR",
            ]:
                raise ValueError(
                    "time_unit must be one of QUARTER_OF_AN_HOUR, HOUR, DAY, WEEK, MONTH, YEAR"
                )
        url = f"{self.url}/site/{self.site_id}/meters"
        if meters:
            meters = ",".join(meters)
        query_params = {
            "api_key": self.api_key,
            "startTime": start_time.strftime("%Y-%m-%d %H:%M:%S"),
            "endTime": end_time.strftime("%Y-%m-%d %H:%M:%S"),
        }
        if time_unit:
            query_params["timeUnit"] = time_unit
        if meters:
            query_params["meters"] = meters

        response = self.session.get(url, params=query_params)
        if response.ok:
            data = response.json()
            return MeterDataResponse(**data)
        logger.error("Failed to retrieve meter data")
        return None



import os

from urllib3.util import Retry
from requests import Session
from requests.adapters import HTTPAdapter
from loguru import logger
from typing import Union

from .site import list_sites


class SolarEdgeClient:
    url = "https://monitoringapi.solaredge.com"
    session = Session()

    def __init__(self, api_key: str, site_id: Union[int, None] = None):
        self.api_key = api_key
        self.__init_session()
        if not site_id:
            site_id = self.get_site()

        self.site_id = site_id

    def __init_session(self):
        retries = Retry(
            total=3,
            backoff_factor=0.1,
            status_forcelist=[429, 502, 503, 504],
            allowed_methods={"GET"},
        )
        self.session.headers.update({"Accept": "application/json"})
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

    def get_site(self) -> int:
        """Get site ID from SolarEdge API or from disk"""
        if os.path.isfile("site_id.txt"):
            with open("site_id.txt", "r") as f:
                return int(f.read())
        site_response = list_sites(self)
        if site_response:
            if site_response.count > 1:
                logger.warning(
                    "More than one site found; Only first site is currently supported"
                )
                # TODO: Extend for multiple sites
            with open("site_id.txt", "w") as f:
                f.write(str(site_response.site[0].id))
            return site_response.site[0].id
        logger.error("Failed to retrieve site ID")
        raise AttributeError("Site ID must be defined and valid")



from typing import List, Optional
from pydantic import BaseModel, NaiveDatetime


class TelemetryData(BaseModel):
    date: NaiveDatetime
    totalActivePower: Optional[float]
    dcVoltage: Optional[float]
    groundFaultResistance: Optional[float] = None
    powerLimit: float
    totalEnergy: float
    temperature: Optional[float] = None
    inverterMode: str
    operationMode: int
    vL1To2: Optional[float] = None
    vL2To3: Optional[float] = None
    vL3To1: Optional[float] = None

    class L1Data(BaseModel):
        acCurrent: float
        acVoltage: float
        acFrequency: float
        apparentPower: float
        activePower: float
        reactivePower: float
        cosPhi: float

    class L2Data(BaseModel):
        acCurrent: float
        acVoltage: float
        acFrequency: float
        apparentPower: float
        activePower: float
        reactivePower: float
        cosPhi: float

    class L3Data(BaseModel):
        acCurrent: float
        acVoltage: float
        acFrequency: float
        apparentPower: float
        activePower: float
        reactivePower: float
        cosPhi: float

    threePhaseInverterTelemetry: Optional[List[dict]] = None


class TelemetryResponse(BaseModel):
    count: int
    telemetries: List[TelemetryData]


class ChangeLog(BaseModel):
    serialNumber: str
    partNumber: Optional[str]
    date: str


class ChangeLogResponse(BaseModel):
    count: int
    list: ChangeLog


class Meter(BaseModel):
    name: str
    manufacturer: str
    model: str
    firmwareVersion: str
    connectedSolaredgeDeviceSN: str
    type: str
    form: str


class Sensor(BaseModel):
    connectedSolaredgeDeviceSN: str
    id: str
    connectedTo: str
    category: str
    type: str


class Gateway(BaseModel):
    name: str
    firmwareVersion: str
    SN: str


class Battery(BaseModel):
    name: str
    manufacturer: str
    model: str
    firmwareVersion: str
    connectedInverterSn: str
    nameplateCapacity: float
    SN: str


class InventoryInverter(BaseModel):
    model: str
    firmwareVersion: str
    SN: str
    connectedOptimizers: int


class Inventory(BaseModel):
    meters: List[Meter]
    sensors: List[Sensor]
    gateways: List[Gateway]
    batteries: List[Battery]
    inverters: List[InventoryInverter]


class InventoryResponse(BaseModel):
    Inventory: Inventory


class Inverter(BaseModel):
    name: str
    manufacturer: str
    model: str
    serialNumber: str


class ComponentsListResponse(BaseModel):
    count: int
    list: List[Inverter]


class Location(BaseModel):
    country: str
    state: Optional[str] = None
    city: str
    address: str
    address2: Optional[str] = None
    zip: str
    timeZone: str


class Site(BaseModel):
    id: int
    name: str
    accountId: int
    status: str
    peakPower: float
    currency: str
    installationDate: str
    ptoDate: Optional[str]
    notes: str
    type: str
    location: Location
    alertQuantity: Optional[int] = None
    alertSeverity: Optional[str] = None
    uris: dict
    publicSettings: dict


class SitesResponse(BaseModel):
    count: int
    site: List[Site]


class MeterEnergyValue(BaseModel):
    date: str
    value: int


class MeterEnergyMeter(BaseModel):
    meterSerialNumber: str
    connectedSolaredgeDeviceSN: str
    model: str
    meterType: str
    values: List[MeterEnergyValue]


class MeterEnergyDetails(BaseModel):
    timeUnit: str
    unit: str
    meters: List[MeterEnergyMeter]


class MeterDataResponse(BaseModel):
    meterEnergyDetails: MeterEnergyDetails



from datetime import datetime
import influxdb_client
from influxdb_client.client.write_api import SYNCHRONOUS
from typing import Union, List, Tuple, Optional


class InfluxDBClient:
    def __init__(self, path: str):
        self.client = influxdb_client.InfluxDBClient.from_config_file(path)
        self.write_api = self.client.write_api(write_options=SYNCHRONOUS)

    def write(self, data: str, bucket: str) -> None:
        """Write data to InfluxDB"""
        self.write_api.write(bucket=bucket, record=data)

    def convert_to_point(
        self,
        time: datetime,
        measurement: str,
        field_values: List[Tuple[str, Union[str, int, float]]],
        tags: Optional[List[Tuple[str, str]]] = None,
    ) -> influxdb_client.Point:
        """Convert a dictionary to an InfluxDB point"""
        if tags is None:
            tags = []
        point = influxdb_client.Point(measurement)
        for fieldvalue in field_values:
            point = point.field(fieldvalue[0], fieldvalue[1])
        for tag in tags:
            point = point.tag(tag[0], tag[1])
        point.time(time, write_precision="ms")
        return point



import os
from suntime import Sun, SunTimeException
from datetime import datetime, timedelta, timezone
import pytz
from loguru import logger

from solaredge_influxdb.solaredge import Equipment, Meter
from solaredge_influxdb.influxdb import InfluxDBClient


def app(
    config_path: str = "./solaredge_influxdb/config.toml",
    latitude: float = os.getenv("LATITUDE", 52.3676),
    longitude: float = os.getenv("LONGITUDE", 4.9041),
    api_key: str = os.getenv("API_KEY"),
    additional_time_window: int = 60,
    timezone_str: str = "Europe/Amsterdam",
):
    sun = Sun(latitude, longitude)
    current_time = datetime.now(timezone.utc)
    logger.debug(f"Current time: {current_time}")
    try:
        sunrise = sun.get_sunrise_time()
        sunset = sun.get_sunset_time()
        logger.debug(f"Sunrise: {sunrise}, Sunset: {sunset}")

    except SunTimeException:
        logger.error("Failed to retrieve sunrise/sunset times")
        raise Exception(
            "Application requires sunset and sunrise times to prevent unnecessary API calls"
        )
    InfluxClient = InfluxDBClient(config_path)
    _timezone = pytz.timezone(timezone_str)

    if (
        sunrise - timedelta(minutes=additional_time_window)
        < current_time
        < sunset + timedelta(minutes=additional_time_window)
    ):
        logger.debug("The Sun is shining bright, let's collect some data!")
        EquipmentClient = Equipment(api_key)
        MeterClient = Meter(api_key)
        current_time = current_time.astimezone(_timezone)
        for inverter in EquipmentClient.inverters:
            tech_data = EquipmentClient.get_technical_data(
                inverter.serialNumber,
                current_time - timedelta(minutes=15),
                current_time,
            )
            if tech_data is None:
                logger.error("Failed to retrieve technical data")
                continue

            for telemetry in tech_data.telemetries:
                tags = [
                    ("serial_number", inverter.serialNumber),
                    ("model", inverter.model),
                    ("operation_mode", telemetry.operationMode),
                    ("inverter_mode", telemetry.inverterMode),
                ]
                telemetry_date = _timezone.localize(telemetry.date)
                energy_point = InfluxClient.convert_to_point(
                    telemetry_date,
                    "solar",
                    [
                        ("total_energy", telemetry.totalEnergy / 1000),
                    ],
                    tags,
                )
                power_point = InfluxClient.convert_to_point(
                    telemetry_date,
                    "solar",
                    [
                        ("ac_power", telemetry.totalActivePower / 1000),
                    ],
                    tags,
                )
                voltage_point = InfluxClient.convert_to_point(
                    telemetry_date,
                    "solar",
                    [
                        ("dc_voltage", telemetry.dcVoltage),
                        ("voltage_l1_to_2", telemetry.vL1To2),
                        ("voltage_l2_to_3", telemetry.vL2To3),
                        ("voltage_l3_to_1", telemetry.vL3To1),
                    ],
                    tags,
                )
                InfluxClient.write(energy_point, "energy")
                InfluxClient.write(power_point, "energy_flow")
                InfluxClient.write(voltage_point, "voltage_current")
    else:
        logger.info("It's dark outside, no need to collect data")
